/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
//this progm reads in the trie and a stacgm file, and obtains the postings -> (blkno, pathno, offset) for each term in the trie
//it runs it just like composition; we compute states reache after a block, and pass on states that are non-zero, so as to continue later
//note that as per the algo, we need to inculde the 0 state after each char -> substring match
//also, within a line, we need only transn from one state to another if matched and copy the aug info (only the offset)
#include "indexing/indexgenUtils.h"

int main(int argc, char** argv) {
	if(argc < 5) {
		cout << "Usage: indexgen <dict preftrie .txt file> <# states in preftrie> <.stac file> <.index file>\n";
		return 1;
	}
	ifstream dict(argv[1], ios::in);
	ifstream fin(argv[3], ios::in);
	if(!dict.is_open()) {
		cout << "Error: Cannot open input file " << argv[1] << endl;
		return 1;
	}
	if(!fin.is_open()) {
		cout << "Error: Cannot open output file " << argv[3] << endl;
		return 1;
	}
	int NUMST = atoi(argv[2]);
	
	int M, K, numblks;
	string line1;
	getline(fin, line1);
	stringstream line1str(line1);
	line1str >> M >> K;
	string line2;
	getline(fin, line2);
	stringstream line2str(line2);
	line2str >> numblks;
	map <int, vector <int> > stacgraph;		//map from blkno to list of children blknos
	for(int i=0; i<numblks; i++) {
		string gphline;
		getline(fin, gphline);
		char* linetok = strtok((char*)gphline.c_str(), "\t");
		int thisvx = atoi(linetok);		//the head int is blk num
		vector <int> emptyv;
		stacgraph[thisvx] = emptyv;
		linetok = strtok(NULL, "\t");
		while((linetok != NULL)&&(strcmp(linetok, "") != 0)) {
			stacgraph[thisvx].push_back(atoi(linetok));
			linetok = strtok(NULL, "\t");
		}
	}
	map <int, vector <string> > allpaths;		//map from blkno to list of blk's paths
	int totpaths = 0;
	for(int b=0; b<numblks; b++) {
		int blknum, tmp, numpaths;
		string linet;
		getline(fin, linet);
		stringstream linetstr(linet);
		linetstr >> blknum >> tmp >> numpaths;	//the blknum need NOT be b!
		for(int p=0; p<numpaths; p++) {		//the pathnum is p
			string path, pathwt;
			getline(fin, path);
			allpaths[blknum].push_back(toLower(path));	//Make it case-insensitive
			getline(fin, pathwt);
			totpaths++;
		}
	}
	
	int *preftrie = (int*) malloc(sizeof(int) * NUMST * NUMCHAR);
	memset(preftrie, '\0', sizeof(int) * NUMST * NUMCHAR);	//if entry is 0, it means REJECT transn; we cannot transn to 0 from other states
	map <int, string> finals;		//list of finalstate to dict term mapping -> to be used later
	while(!dict.eof()) {
		string trans;			//could be: from\tto\til\tol or finalst alone
		getline(dict, trans);
		if(trans.compare("") == 0)
			continue;
		char* linetok = strtok((char*)trans.c_str(), "\t");
		int thisvx = atoi(linetok);		//the head int is blk num
		linetok = strtok(NULL, "\t");
		map <char, int> emptyv;
		if(linetok == NULL)
			finals[thisvx] = "";
		else {
			int nxst = atoi(linetok);
			linetok = strtok(NULL, "\t");
			int ch = atoi(linetok);
			preftrie[thisvx*NUMCHAR + (ch - 32)] = nxst;
		}
	}
	//invoke forward dfs on the preftrie, and keep aggr the string, and when we reach a final state, obtain mapping from finalst to term
	fwdDFS(preftrie, &finals, 0, "");	//start from state 0 with empty str
	
	//now we have the trie and terms-finst map in memory; next, we need to compute state trsansn per block this we can do in a dfs with memoizn
	int numchildren[numblks];
	int visitcnt[numblks];
	map <int, vector <int> > stacgraphrev;		//the graph of blks reversed - to be used later
	for(int i=0; i<numblks; i++) {
		vector <int> emptyv;
		stacgraphrev[i] = emptyv;		//will update this later
		visitcnt[i] = 0;
		numchildren[i] = stacgraph[i].size();
	}
	//get the stacgraphrev; also the outdegrees of blknodes -> to get the end nodes
	vector <int> endblks;
	for(map <int, vector <int> >::iterator iter = stacgraph.begin(); iter != stacgraph.end(); iter++) {
		for(vector <int>::iterator siter = iter->second.begin(); siter != iter->second.end(); siter++)
			stacgraphrev[*siter].push_back(iter->first);		//add rev edge
		if(iter->second.size() == 0)
			endblks.push_back(iter->first);
	}
	
	augstates memoaugstates[numblks];
	int memoflag[numblks];
	for(int y=0; y<numblks; y++)
		memoflag[y] = -1;
	augstates emissions;

	for(vector <int>::iterator siter = endblks.begin(); siter != endblks.end(); siter++)
		augstates finalret = recEmit(&allpaths, &stacgraphrev, preftrie, &finals, *siter, memoaugstates, memoflag, &emissions, numchildren, visitcnt);
	
	ofstream foutindex(argv[4], ios::out);
	if(!foutindex.is_open()) {
		cout << "Error: Cannot open index file " << argv[4] << endl;
		return 1;
	}
	//.index file schema is: <term> \t : \t numpostings blkno pathno off blkno pathno off ... \n
	for(map <int, set <location> >::iterator iter = emissions.begin(); iter != emissions.end(); iter++) {
		foutindex << finals[iter->first] << "\t:\t" << iter->second.size();
		for(set <location>::iterator liter = iter->second.begin(); liter != iter->second.end(); liter++)
			foutindex << " " << liter->first << " " << liter->second.first << " " << liter->second.second;
		foutindex << endl;
	}
return 0;
}
