/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "libraries/fstUtils.h"
#include <fstream>

//mbin format:
//<numstates><startstatenum><finalstatenum><numtrans>
//for each transition:<fromstatenum><tostatenum><numarcs>
//foreach arc:<ilabel><olabel><weight>
int main(int argc, char** argv) {
	if(argc < 4) {
		cout << "Usage: fstmbin <.fst file> <.mbin file> <flag: 0 for fst to mbin | 1 for mbin to fst>\n";
		return 0;
	}
	int flag = atoi(argv[3]);
	if(flag == 1) {		//convert mbin to fst format and write to outstream; prov info is default included
		ifstream inmbin(argv[2], ios::in|ios::binary);
		StdVectorFst nfst;
		mbintofst(&nfst, inmbin);	//this is the istream version
		nfst.Write(argv[1]);
		inmbin.close();
	}
	else {	//convert fst to mbin format and write to outstream
		StdVectorFst* ifst = StdVectorFst::Read(argv[1]);
		ofstream fout(argv[2], ios::out|ios::binary);
		TopSort(ifst);	//sort the fst nodes topologically
		StdVectorFst newfst = augmentFst(ifst); //sanitize probabilities; augment with edge provenance
		fsttombin(&newfst, fout);
		fout.close();
	}	
return 0;
}
